import {
  OllamaRequest,
  OllamaResponse,
  NewStatusBody,
  Notification,
  OllamaConfigOptions,
} from "../types.js";
import striptags from "striptags";
import { PrismaClient } from "../generated/prisma/client.js";

const prisma = new PrismaClient();

const envConfig = {
  pleromaInstanceUrl: process.env.PLEROMA_INSTANCE_URL || "",
  pleromaInstanceDomain: process.env.PLEROMA_INSTANCE_DOMAIN || "",
  whitelistOnly: process.env.ONLY_WHITELIST === "true" ? true : false || "true",
  whitelistedDomains: process.env.WHITELISTED_DOMAINS
    ? process.env.WHITELISTED_DOMAINS.split(",")
    : [process.env.PLEROMA_INSTANCE_DOMAIN],
  ollamaUrl: process.env.OLLAMA_URL || "",
  ollamaSystemPrompt:
    process.env.OLLAMA_SYSTEM_PROMPT ||
    "You are a helpful AI assistant. Answer all questions concisely.",
  ollamaModel: process.env.OLLAMA_MODEL || "",
  fetchInterval: process.env.FETCH_INTERVAL
    ? parseInt(process.env.FETCH_INTERVAL)
    : 15000,
  bearerToken: process.env.INSTANCE_BEARER_TOKEN || "",
};

const ollamaConfig: OllamaConfigOptions = {
  temperature: 0.6,
  num_predict: 750,
};

const getNotifications = async () => {
  const { bearerToken, pleromaInstanceUrl } = envConfig;
  try {
    const request = await fetch(
      `${pleromaInstanceUrl}/api/v1/notifications?types[]=mention`,
      {
        method: "GET",
        headers: {
          Authorization: `Bearer ${bearerToken}`,
        },
      }
    );

    const notifications: Notification[] = await request.json();

    return notifications;
  } catch (error: any) {
    throw new Error(error.message);
  }
};

const storeUserData = async (notification: Notification): Promise<void> => {
  try {
    await prisma.user.upsert({
      where: { userFqn: notification.status.account.fqn },
      update: {
        lastRespondedTo: new Date(Date.now()),
      },
      create: {
        userFqn: notification.status.account.fqn,
        lastRespondedTo: new Date(Date.now()),
      },
    });
  } catch (error: any) {
    throw new Error(error.message);
  }
};

const alreadyRespondedTo = async (
  notification: Notification
): Promise<boolean> => {
  try {
    const duplicate = await prisma.response.findFirst({
      where: { pleromaNotificationId: notification.id, isProcessing: true },
    });
    if (duplicate) {
      return true;
    }
    return false;
  } catch (error: any) {
    throw new Error(error.message);
  }
};

const storePromptData = async (
  notification: Notification,
  ollamaResponseBody: OllamaResponse
) => {
  try {
    await prisma.response.updateMany({
      where: { pleromaNotificationId: notification.id },
      data: {
        response: ollamaResponseBody.response,
        request: trimInputData(notification.status.content),
        to: notification.account.fqn,
        isProcessing: false,
      },
    });
  } catch (error: any) {
    throw new Error(error.message);
  }
};

const trimInputData = (input: string): string => {
  const strippedInput = striptags(input);
  const split = strippedInput.split(" ");
  const promptStringIndex = split.indexOf("!prompt");
  split.splice(promptStringIndex, 1);
  return split.join(" "); // returns everything after the !prompt
};

const recordPendingResponse = async (notification: Notification) => {
  try {
    await prisma.response.create({
      data: {
        pleromaNotificationId: notification.id,
      },
    });
  } catch (error: any) {
    throw new Error(error.message);
  }
};

const isFromWhitelistedDomain = (fqn: string): boolean => {
  try {
    const domain = fqn.split("@")[1];
    if (envConfig.whitelistedDomains.includes(domain)) {
      return true;
    }
    return false;
  } catch (error: any) {
    console.error(`Error with domain check: ${error.message}`);
    return false;
  }
};

const generateOllamaRequest = async (
  notification: Notification
): Promise<OllamaResponse | undefined> => {
  const { whitelistOnly, ollamaModel, ollamaSystemPrompt, ollamaUrl } =
    envConfig;
  try {
    if (
      striptags(notification.status.content).includes("!prompt") &&
      !notification.status.account.bot && // sanity check, sort of
      notification.type === "mention"
    ) {
      if (
        whitelistOnly &&
        !isFromWhitelistedDomain(notification.status.account.fqn)
      ) {
        return;
      }
      if (await alreadyRespondedTo(notification)) {
        return;
      }
      await recordPendingResponse(notification);
      await storeUserData(notification);
      const ollamaRequestBody: OllamaRequest = {
        model: ollamaModel,
        system: ollamaSystemPrompt,
        prompt: `@${notification.status.account.fqn} says: ${trimInputData(
          notification.status.content
        )}`,
        stream: false,
        options: ollamaConfig,
      };
      const response = await fetch(`${ollamaUrl}/api/generate`, {
        method: "POST",
        body: JSON.stringify(ollamaRequestBody),
      });
      const ollamaResponse: OllamaResponse = await response.json();
      await storePromptData(notification, ollamaResponse);
      return ollamaResponse;
    }
  } catch (error: any) {
    throw new Error(error.message);
  }
};

const postReplyToStatus = async (
  notification: Notification,
  ollamaResponseBody: OllamaResponse
) => {
  const { pleromaInstanceUrl, bearerToken } = envConfig;
  try {
    let mentions: string[];
    const statusBody: NewStatusBody = {
      content_type: "text/markdown",
      status: ollamaResponseBody.response,
      in_reply_to_id: notification.status.id,
    };
    if (
      notification.status.mentions &&
      notification.status.mentions.length > 0
    ) {
      mentions = notification.status.mentions.map((mention) => {
        return mention.acct;
      });
      statusBody.to = mentions;
    }

    const response = await fetch(`${pleromaInstanceUrl}/api/v1/statuses`, {
      method: "POST",
      headers: {
        Authorization: `Bearer ${bearerToken}`,
        "Content-Type": "application/json",
      },
      body: JSON.stringify(statusBody),
    });

    if (!response.ok) {
      throw new Error(`New status request failed: ${response.statusText}`);
    }

    await deleteNotification(notification);
  } catch (error: any) {
    throw new Error(error.message);
  }
};

const deleteNotification = async (notification: Notification) => {
  const { pleromaInstanceUrl, bearerToken } = envConfig;
  try {
    if (!notification.id) {
      return;
    }
    await prisma.response.updateMany({
      // this is probably not the best way to do this, but since we may have duplicate notifications, we have to update all of them - probably won't scale (lmao)
      where: { pleromaNotificationId: notification.id },
      data: { isProcessing: false },
    });
    const response = await fetch(
      `${pleromaInstanceUrl}/api/v1/notifications/${notification.id}/dismiss`,
      {
        method: "POST",
        headers: {
          Authorization: `Bearer ${bearerToken}`,
        },
      }
    );
    if (!response.ok) {
      console.error(
        `Could not delete notification ID: ${notification.id}\nReason: ${response.status} - ${response.statusText}`
      );
    }
  } catch (error: any) {
    throw new Error(error.message);
  }
};

let notifications = [];
const beginFetchCycle = async () => {
  setInterval(async () => {
    notifications = await getNotifications();
    if (notifications.length > 0) {
      await Promise.all(
        notifications.map(async (notification) => {
          try {
            const ollamaResponse = await generateOllamaRequest(notification);
            if (ollamaResponse) {
              postReplyToStatus(notification, ollamaResponse);
            }
          } catch (error: any) {
            throw new Error(error.message);
          }
        })
      );
    }
  }, envConfig.fetchInterval); // lower intervals may cause the bot to respond multiple times to the same message, but we try to mitigate this with the deleteNotification function
};

console.log(
  `Fetching notifications from ${envConfig.pleromaInstanceDomain}, every ${
    envConfig.fetchInterval / 1000
  } seconds.`
);
await beginFetchCycle();
